
//Program K15 DeleteLeaf2
//This program generates a random tree and 
//deletes the last inserted leaf. 

//We apply two deleting rules:
//antiRule 1:
//(An,Fn)i  -> Ai
//antiRule 2:
//((S)n,Fn)i -> (S)i
//where A and F are leaves, n and i are  inner nodes,
//(S) is a complex structure.
//These rules are inverse to inserting rules.

//Both rules are implemented.



import java.util.Random;


public class DeleteLeaf2
{

 //Link's notation for trees
 private static   String   parent = 
		 "((Aa,Ba),Ca)";
 private static  String leaves =  
		 "ABCDEFGHIJKLMNOPRSTUVWXYZ";
 private static String internalNodes = 
		 "abcdefghijklmnopqrstuvwxyz";
 //Next leaf to be grafted	
 private static  int nextLeafIndex = 2;
 //Next internal node to be inserted
 private static  int nextIntNodeIndex = 0;
 
 private static  Random r = new Random();
 
 private static boolean print1 = false;
 private static boolean print2 = true; 
 
 //After deleting the last inserted leaf, 
 //this must result:
 private static  String expected = "";

 //Detect internal nodes, different than leaves
 private static boolean charIsInternalNode(char c)
 {
  String s = "" + c;	 
  if (internalNodes.contains(s)) return true;
  else return false;
 }
 
//Simplification from Link notation to Newick's notation:
 // (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)
private static String simplify(String newTree)
{
 String tree = "";
//length of string newTree
int l = newTree.length();
 for(int i = 0; i < l; i++ )
  {
   //newTree is scanned char by char	 
  char c = newTree.charAt(i);
  //Is c an internal node to be deleted?
  if (charIsInternalNode(c));
   else //add char c to output
    tree = tree + c;
  }
 return tree;	 
}
 
//Rule 2 is used to generate a new tree:
//(S)I -> ((S)N,FN)I.
//S = subtree, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string parentTree
private static String useRule2(String parentTree, int i, 
        char F, char N)
{
 String newTree = "";
 char I = parentTree.charAt(i);
 if (print1)
 System.out.println("i = " + i + " I = " + I);
 //Subtree S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = i;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
   String a = "" + parentTree.charAt(position);
   
   if (print1)
   System.out.println("a = " + a + " p = " + pr);
   if (a.equals(pr) ) 
   {
	counter++;
	S = a + S;
   }
   else go = false;
  }//End of while
 if (print1)
 System.out.println("counter = " + counter);
 //The complete subtree is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String a = "" + parentTree.charAt(position);
   if (a.equals(pr))   counter++;
   if (a.equals(pl))  counter--;
   S = a + S;
   if (print1)
   System.out.println("Subtree = " + S);
   position = position -1;
   }
 //Replacement rule 2:
 //((S)n,Fn)i -> (S)i
//where   F are leaves, n and i are  inner nodes,
//(S) is a complex structure. 
 String SnFni	 =   S  + I; 
 String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
 //Rule 1 is executed
 newTree = parentTree.replace(SnFni, replacement);		
 //If I links to a complex structure
 return newTree;	 
}

//From parentTree, a tree in Newick's notation,
//a Child  tree is generated by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String parentTree) 
{
	
 nextLeafIndex++;
 char nextLeaf= leaves.charAt(nextLeafIndex);
 //Next internal node to be inserted
 nextIntNodeIndex++;
 char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
	
	
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new tree generated by replacement
 String newTree = "";
 //length of parent tree
 int l = parentTree.length();
 //counter of links = 0;
 int counter = 0;
 //Links are counted
 for(int i = 0; i < l; i++ )
  {
   //parentTree is scanned char by char	 
   char I = parentTree.charAt(i);
	//Is I an internal node that denotes a link?
	if (charIsInternalNode(I)) counter++;	
  }
 //A link is chosen 
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
  {
    //parentTree is scanned char by char	 
	char I = parentTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = parentTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(parentTree, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTree = parentTree.replace(link, replacement);		
		 }
		System.out.println(newTree + " -> " 
		                       + simplify(newTree));			
 
	  
		 i = l;
	   }
	  counter++;	
	} 
  } 
 return newTree;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply the antirule1:
//(An,Fn)i  -> Ai
private static String  antiRule1(String tree, 
		int positionLeaf, int positionNode)
{
 System.out.println("AntiRule1");
 int beginIndex = positionLeaf - 4;
 int endIndex = positionNode + 3;
 String AnFni = tree.substring(beginIndex, endIndex);
 System.out.println("Anfni = " + AnFni);
 String Ai = "" + tree.charAt(beginIndex+1)
                + tree.charAt(endIndex-1);
 System.out.println("Ai = " + Ai);
 tree = tree.replace(AnFni, Ai);		
	return tree;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply the antirule2:
//((S)n,Fn)i -> (S)i
//positionNode refers to the last appearing of node n
private static String  antiRule2(String tree, 
		int positionLeaf, int positionNode)
{
	System.out.println("AntiRule2");
 //Leaf to be deleted	
 String leaf = "" + leaves.charAt(nextLeafIndex);
 System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
	            + internalNodes.charAt(nextIntNodeIndex);	
 int firstPosition = tree.indexOf(innerNode);
 String newTree = "";
 char i = tree.charAt(positionNode+2);
 if (print2)
 System.out.println("FirstPosition = " + 
                      firstPosition + " i = " + i);
 //Subtree S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = firstPosition;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
    String as = "" + tree.charAt(position);
   
   if (print2)
   System.out.println("a = " + as + " p = " + pr);
   if (as.equals(pr) ) 
   {
	counter++;
	S = as + S;
   }
   else go = false;
  }//End of while
 if (print2)
 System.out.println("counter = " + counter);
 //The complete subtree is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String as = "" + tree.charAt(position);
   if (as.equals(pr))   counter++;
   if (as.equals(pl))  counter--;
   S = as + S;
   if (print2)
   System.out.println("Subtree = " + S);
   position = position -1;
   }
  String n = innerNode;
 //Replacement antiRule 2:
 //((S)n,Fn)i -> (S)i
 //where A and F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =  "("+ S  + n + "," + leaf + n + ")" + i ; 
 String	Si = S + i;
 if (print2)
  System.out.println("\nSnFni = " + SnFni + "\n Si = " + Si);
 //Rule 1 is executed
 newTree = tree.replace(SnFni, Si);		
 
  
 
  
 return newTree;
}



//The last inserted  leaf is deleted from tree.

private static String deleteLeaf(String tree) 
{
  //Leaf to be deleted	
  String leaf = "" + leaves.charAt(nextLeafIndex);
  System.out.println("\nLeaf to be deleted = " + leaf);
  //Accompanying inner node  
  String innerNode = "" 
            + internalNodes.charAt(nextIntNodeIndex);
  System.out.println("Accompanying inner  node = " 
                    + innerNode);		
  //new tree = output
  String newTree = "";
  //length of  tree as string
  int l = tree.length();
  //Last positions are found
  int positionLeaf = -1;
  int positionNode = -1;  
  //tree is scanned char by char
  for(int i = 0; i < l; i++ )
   {
	 String c = "" + tree.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
   }
  System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
  System.out.println("Position of " + innerNode + 
          " = " + positionNode);
  int firstPosition = tree.indexOf(innerNode);
  //Check for the char before firstPosition	
  char a = tree.charAt(firstPosition - 1);
  //If I links to a complex structure
  if (a == ')')
  newTree = antiRule2(tree,positionLeaf,positionNode);
  else  //Apply antiRule1 for simple cases
	  newTree = antiRule1(tree,positionLeaf,positionNode);
  return newTree;	
}


public static void main(String[] args) 
{
 //A tree is generated	
 System.out.println("Parent tree = \n" + parent
		 + " -> "    + simplify(parent)); 
 //Number of leaves
 int nLeaves = 13; 
 for(int i = 0; i < nLeaves; i++ )	 
  {
   parent = generateChild(parent);
   if (i == nLeaves -2) expected = parent; 
  }
 //Last inserted leaf is deleted.
 String answer = deleteLeaf(parent);
 System.out.println("\nExpected and actual answer = \n" 
    + expected + "\n" + answer);
 if (answer.equals(expected))
	  System.out.println("Answer is correct.");
  else
	  System.out.println("\nAnswer is wrong!"); 
}

}//End of main class K15 DeleteLeaf2