
//Program K28 SubTopol2
/*
 * This program list all possible subtopologies
 * of a given topology.
 * 
 * We begin with a left parenthesis "(" and cumulate 
 * symbols until the number of
 * right and left parentheses get balanced.
 * 
 * All left parenthesis are investigated.
 * 
 * Subtopologies are packed for future use.
 * 
 */

import java.util.Random;

public class SubTopol2
{

	//Link's notation for trees
	 private static   String   parent = 
			 "((Aa,Ba),Ca)";
	 private static  String leaves =  
			 "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	 private static String internalNodes = 
			 "abcdefghijklmnopqrstuvwxyz";
	 //Next leaf to be grafted	
	 private static  int nextLeafIndex = 2;
	 //Next internal node to be inserted
	 private static  int nextIntNodeIndex = 0;
	 
	 //Leaves that remain in the tree
	 private static  String presentLeaves =   "ABC";
	 private static String presentNodes =  "a";

	 
	 private static  Random r = new Random();
	 
	 private static boolean print1 = false;
	 private static boolean print2 = false; 
	
	 

//====== A random topology is generated ============= 

//Detect internal nodes, different than leaves
private static boolean charIsInternalNode(char c)
 {
  String s = "" + c;	 
  if (internalNodes.contains(s)) return true;
   else return false;
 }


//Simplification from Link notation to Newick's notation:
// (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)	 	 
private static String simplify(String newTree)
 {
  String tree = "";
  //length of string newTree
  int l = newTree.length();
  for(int i = 0; i < l; i++ )
	  {
	   //newTree is scanned char by char	 
	  char c = newTree.charAt(i);
	  //Is c an internal node to be deleted?
	  if (charIsInternalNode(c));
	   else //add char c to output
	    tree = tree + c;
	  }
	 return tree;	 
 }

 //Rule 2 is used to generate a new tree:
 //(S)I -> ((S)N,FN)I.
 //S = subtree, I = internal node, F = new leaf, 
 //N = new internal node.
 //Rule 2 grafts new branch FN to link (S)I
 //i is the position of char I in string parentTree
 private static String useRule2(String parentTree, int i, 
	        char F, char N)
 {
  String newTree = "";
  char I = parentTree.charAt(i);
  if (print1)
  System.out.println("i = " + i + " I = " + I);
  //Subtree S of link (S)I must be read 
  String S = "";
  //Counter of left parenthesis
  int counter = 0;
  String pl = "" + '(';
  String pr = "" + ')';
  //Position to be scanned
  int position = i;
  boolean go = true;
  //All right parentheses to the near left of I are captured
  while (go)
	  {
	   position = position -1;
	   String a = "" + parentTree.charAt(position);
	   
	   if (print1)
	   System.out.println("a = " + a + " p = " + pr);
	   if (a.equals(pr) ) 
	   {
		counter++;
		S = a + S;
	   }
	   else go = false;
	  }//End of while
  if (print1)
	 System.out.println("counter = " + counter);
  //The complete subtree is captured:
  //Parentheses must balanced
  while (counter > 0)
	  {   
	   String a = "" + parentTree.charAt(position);
	   if (a.equals(pr))   counter++;
	   if (a.equals(pl))  counter--;
	   S = a + S;
	   if (print1)
	   System.out.println("Subtree = " + S);
	   position = position -1;
	   }
  //Replacement rule 2:
  //((S)n,Fn)i -> (S)i
  //where   F are leaves, n and i are  inner nodes,
  //(S) is a complex structure. 
  String SnFni	 =   S  + I; 
  String	replacement =  "("+ S +   N + "," 
			          + F + N+ ")" + I;
  //Rule 1 is executed
  newTree = parentTree.replace(SnFni, replacement);		
  //If I links to a complex structure
  return newTree;	 	
 }

	
//Input: parentTree, a tree in Newick's notation.
//Output : newTree =  parent tree + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String parentTree) 
 {
  nextLeafIndex++;
  char nextLeaf = leaves.charAt(nextLeafIndex);
  presentLeaves = presentLeaves + nextLeaf;
  //Next internal node to be inserted
  nextIntNodeIndex++;
  char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
  presentNodes = presentNodes + nextIntNode;
		
  //link to be replaced
  String link = "";
  //String link is replaced by what:
  String replacement = "";
  //new tree generated by replacement
  String newTree = "";
  //length of parent tree
  int l = parentTree.length();
  //counter of links = 0;
  int counter = 0;
  //Links are counted
  for(int i = 0; i < l; i++ )
	  {
	   //parentTree is scanned char by char	 
	   char I = parentTree.charAt(i);
		//Is I an internal node that denotes a link?
		if (charIsInternalNode(I)) counter++;	
	  }
  //A link is chosen 
  int k = r.nextInt(counter);
  //System.out.println("Random number = " + k);
  //new branch is grafted at k
  counter = 0;  
  for(int i = 0; i < l; i++ )
   {
	//parentTree is scanned char by char	 
	char I = parentTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	  {
		//Check for the char before I	
		char a = parentTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(parentTree, i, 
					               nextLeaf, nextIntNode);
		 else //if I links to a leaf, use rule 1
		  {
		   //Link to be replaced  
		   link = "" + a + I; 
		   //Replacement rule 1:
		   //aI -> (aN,FN)I;   N= nextIntNode; 
		   //F = nextLeaf
		  replacement = "(" + a + nextIntNode + "," 
				       + nextLeaf + nextIntNode + ")" + I;
		   //Rule 1 is executed
		   newTree = parentTree.replace(link, replacement);		
		  }
		if (print1)
		 System.out.println(newTree + " -> " 
			                       + simplify(newTree));				 		  
		 i = l;
		}
	   counter++;	
	  } 
	 } 
    return newTree;
   }


//========== Subtopologies are listed ============
	

//Subtopologies of  topology are found
public static String[] subtopologies(String topology)
{
 System.out.println("Subtopologies are found");
 //Counter of left parenthesis
 int counter = 0;
 //Counter of subtopologies
 int counterSubtop = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Length of String   topology
 int m = topology.length();
 String[] subtopologies = new String[m];
 //String topology is scanned char by char
 for(int i = 0; i < m; i++ )
  {	 
   String Sub = "";
   String a = "" + topology.charAt(i);
   int position = i+1;
   if (a.equals(pl))
   {   
	counter++;
	 Sub =  Sub + a;
	//The complete subtopology is captured:
	//Parentheses must balanced
	while (counter > 0)
	 {   
	  String b = "" + topology.charAt(position);
	  if (b.equals(pl))   counter++;
	  if (b.equals(pr))  counter--;
	  Sub =  Sub + b;
	 
	  position++;
	 } 	
	 if (print2)
		   System.out.println(Sub);
   subtopologies[counterSubtop] = Sub; 
   counterSubtop++;
  }
 }	 
 return subtopologies;
}


	
 public static void main(String[] args) 
 {   
   System.out.println("A random topology is generated");
   //Seed
   if (print1)
   System.out.println("\n" + parent
			 + " -> "    + simplify(parent)); 		  	
   //A tree topology is generated	
   String topology = parent;
   //Number of leaves to be inserted
   int nLeavesToBeInserted = 6; 
   for(int i = 0; i < nLeavesToBeInserted; i++ )	 
	   topology = generateChild(topology);
   System.out.println(topology + " -> " 
	          + simplify(topology));
   //Length of String   topology
   int m = topology.length();
   String[] subtopologies = new String[m];
   //Subtopologies are found
   subtopologies = subtopologies(topology);
   for(int i = 0; i < m; i++)
    {
	   if (subtopologies[i] == null);
	   else
	   System.out.println(i + " " + subtopologies[i]);
    }
 }
}//End of main class K28  SubTopol2
