
//Program K31 SubTopol4
/*
 * This program generates two topologies  on the same
 * set of leaves. 
 * All their subtopologies are calculated 
 * and packed in arrays.
 * Sets of leaves and inner nodes of each subtopology
 * are found.
 * 
 *
 */

import java.util.Random;

public class SubTopol4
{

	//Link's notation for trees
	 private static   String   parent = 
			 "((Aa,Ba),Ca)";
	 private static  String leaves =  
			 "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	 private static String internalNodes = 
			 "abcdefghijklmnopqrstuvwxyz";
	 //Next leaf to be grafted	
	 private static  int nextLeafIndex = 2;
	 //Next internal node to be inserted
	 private static  int nextIntNodeIndex = 0;
	
	 //===========Size of the tree=========== 
	 //Number of leaves to be inserted
	 private static  int nLeavesToBeInserted = 20; 
	 //=====================================
	 
	 
	 //Counter of subtopologies
	 private static  int counterSubtop;
	 //Particular number of subtopologies
	 private static  int m1 = 0;
	 private static  int m2 = 0;
	 
	 //Leaves that remain in the tree
	 private static  String presentLeaves =   "ABC";
	 private static String presentNodes =  "a";

	 
	 private static  Random r = new Random();
	 
	 private static boolean print1 = false;
	 private static boolean print2 = false; 
	
	 

//====== A random topology is generated ============= 

//Detect internal nodes, different than leaves
private static boolean charIsInternalNode(char c)
 {
  String s = "" + c;	 
  if (internalNodes.contains(s)) return true;
   else return false;
 }


//Simplification from Link notation to Newick's notation:
// (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)	 	 
private static String simplify(String newTree)
 {
  String tree = "";
  //length of string newTree
  int l = newTree.length();
  for(int i = 0; i < l; i++ )
	  {
	   //newTree is scanned char by char	 
	  char c = newTree.charAt(i);
	  //Is c an internal node to be deleted?
	  if (charIsInternalNode(c));
	   else //add char c to output
	    tree = tree + c;
	  }
	 return tree;	 
 }

 //Rule 2 is used to generate a new tree:
 //(S)I -> ((S)N,FN)I.
 //S = subtree, I = internal node, F = new leaf, 
 //N = new internal node.
 //Rule 2 grafts new branch FN to link (S)I
 //i is the position of char I in string parentTree
 private static String useRule2(String parentTree, 
		 int i,   char F, char N)
 {
  String newTree = "";
  char I = parentTree.charAt(i);
  if (print1)
  System.out.println("i = " + i + " I = " + I);
  //Subtree S of link (S)I must be read 
  String S = "";
  //Counter of left parenthesis
  int counter = 0;
  String pl = "" + '(';
  String pr = "" + ')';
  //Position to be scanned
  int position = i;
  boolean go = true;
  //All right parentheses to the near left 
  //of I are captured
  while (go)
	  {
	   position = position -1;
	   String a = "" + parentTree.charAt(position);
	   
	   if (print1)
	   System.out.println("a = " + a + " p = " + pr);
	   if (a.equals(pr) ) 
	   {
		counter++;
		S = a + S;
	   }
	   else go = false;
	  }//End of while
  if (print1)
	 System.out.println("counter = " + counter);
  //The complete subtree is captured:
  //Parentheses must balanced
  while (counter > 0)
	  {   
	   String a = "" + parentTree.charAt(position);
	   if (a.equals(pr))   counter++;
	   if (a.equals(pl))  counter--;
	   S = a + S;
	   if (print1)
	   System.out.println("Subtree = " + S);
	   position = position -1;
	   }
  //Replacement rule 2:
  //((S)n,Fn)i -> (S)i
  //where   F are leaves, n and i are  inner nodes,
  //(S) is a complex structure. 
  String SnFni	 =   S  + I; 
  String	replacement =  "("+ S +   N + "," 
			          + F + N+ ")" + I;
  //Rule 1 is executed
  newTree = parentTree.replace(SnFni, replacement);		
  //If I links to a complex structure
  return newTree;	 	
 }

	
//Input: parentTree, a tree in Newick's notation.
//Output : newTree =  parent tree + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String parentTree) 
 {
  nextLeafIndex++;
  char nextLeaf = leaves.charAt(nextLeafIndex);
  presentLeaves = presentLeaves + nextLeaf;
  //Next internal node to be inserted
  nextIntNodeIndex++;
  char nextIntNode = 
		  internalNodes.charAt(nextIntNodeIndex);
  presentNodes = presentNodes + nextIntNode;
		
  //link to be replaced
  String link = "";
  //String link is replaced by what:
  String replacement = "";
  //new tree generated by replacement
  String newTree = "";
  //length of parent tree
  int l = parentTree.length();
  //counter of links = 0;
  int counter = 0;
  //Links are counted
  for(int i = 0; i < l; i++ )
	  {
	   //parentTree is scanned char by char	 
	   char I = parentTree.charAt(i);
		//Is I an internal node that denotes a link?
		if (charIsInternalNode(I)) counter++;	
	  }
  //A link is chosen 
  int k = r.nextInt(counter);
  //System.out.println("Random number = " + k);
  //new branch is grafted at k
  counter = 0;  
  for(int i = 0; i < l; i++ )
   {
	//parentTree is scanned char by char	 
	char I = parentTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	  {
		//Check for the char before I	
		char a = parentTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(parentTree, i, 
					         nextLeaf, nextIntNode);
		 else //if I links to a leaf, use rule 1
		  {
		   //Link to be replaced  
		   link = "" + a + I; 
		   //Replacement rule 1:
		   //aI -> (aN,FN)I;   N= nextIntNode; 
		   //F = nextLeaf
		  replacement = "(" + a + nextIntNode + "," 
				       + nextLeaf + nextIntNode + ")" + I;
		   //Rule 1 is executed
		   newTree = parentTree.replace(link, replacement);		
		  }
		if (print1)
		 System.out.println(newTree + " -> " 
			                     + simplify(newTree));				 		  
		 i = l;
		}
	   counter++;	
	  } 
	 } 
    return newTree;
   }


//========== Subtopologies are listed ============
	

//Subtopologies of  topology are found
public static String[] subtopologies(String topology)
{
 if (print1)
	System.out.println("\nSubtopologies are found");
 //Counter of left parenthesis
 int counter = 0;
 //Counter of subtopologies
 counterSubtop = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Length of String   topology
 int m = topology.length();
 String[] subtopologies = new String[m];
 //String topology is scanned char by char
 for(int i = 0; i < m; i++ )
  {	 
   String Sub = "";
   String a = "" + topology.charAt(i);
   int position = i+1;
   if (a.equals(pl))
   {   
	counter++;
	 Sub =  Sub + a;
	//The complete subtopology is captured:
	//Parentheses must balanced
	while (counter > 0)
	 {   
	  String b = "" + topology.charAt(position);
	  if (b.equals(pl))   counter++;
	  if (b.equals(pr))  counter--;
	  Sub =  Sub + b;	 
	  position++;
	 } 	
	 if (print2)
		   System.out.println(Sub);
   subtopologies[counterSubtop] = Sub; 
   counterSubtop++;
  }
 }	 
 return subtopologies;
}




 public static String generateTopology()
 {
	 System.out.println("A random topology is generated");
	   //Seed
	   if (print1)
	   System.out.println("\n" + parent
				 + " -> "    + simplify(parent)); 
		 //Next leaf to be grafted	
		 nextLeafIndex = 2;
		 //Next internal node to be inserted
		 nextIntNodeIndex = 0;
	   //A tree topology is generated	
	   String topology = parent;
	   for(int i = 0; i < nLeavesToBeInserted; i++ )	 
		   topology = generateChild(topology);
	   return topology;
 }
 
 
 
 //Leaves of subtopologies  are found
  public static String[]leaves(int m, String[] subtop)
 {
  if (print1)  
   System.out.println("Leaves are found");
  String[] myLeaves = new String[m];
  for(int i = 0; i < m; i++)
   {
	 String s = subtop[i];
	 if (s == null);
	   else
	   {
	    myLeaves[i] = "";
	    int n = s.length();
	    for(int j = 0; j < n; j++)
	     {
	      String c = "" + subtop[i].charAt(j);
	      if (leaves.contains(c))
	    	  myLeaves[i] = myLeaves[i] + c;
	     }	
	    if (print1)
	    System.out.println(subtop[i] + " "
	                           + myLeaves[i] );
	   }	
    }
  //Chars of Leaves are rewritten in alphabet order
  if (print1)
  System.out.println("Leaves are listed orderly" );
  for(int i = 0; i < m; i++)
  {
	 String s = subtop[i];
	 if (s == null);
	   else
	   {
	     s =   myLeaves[i];
	     myLeaves[i] = "";
	    int n = leaves.length();
	    for(int j = 0; j < n; j++)
	     {
	      String c = "" + leaves.charAt(j);
	      if (s.contains(c))
	    	  myLeaves[i] = myLeaves[i] + c;
	     }	
	    if (print1)
	    System.out.println(subtop[i] + " " + myLeaves[i] );
	   }	
   } 
  return myLeaves;
   }
  
 
  
  //Inner nodes of subtopologies  are found
   public static String[] inner(int m, String[] subtop)
  {
   if (print1)
   System.out.println("\nInternal nodes are found");
   String[] myNodes = new String[m];
   for(int i = 0; i < m; i++)
    {
 	 String s = subtop[i];
 	 if (s == null);
 	   else
 	   {
 		  myNodes[i] = "";
 	    int n = s.length();
 	    for(int j = 0; j < n; j++)
 	     {
 	      String c = "" + subtop[i].charAt(j);
 	      if (internalNodes.contains(c))
 	    	 myNodes[i] = myNodes[i] + c;
 	     }	
 	   if (print1)
 		System.out.println(subtop[i] + " " + myNodes[i] );
 	   }	
     }
   //Chars of Leaves are rewritten in alphabet order
   if (print1)
   System.out.println("Inner nodes are listed orderly" );
   for(int i = 0; i < m; i++)
   {
 	 String s = subtop[i];
 	 if (s == null);
 	   else
 	   {
 	     s =   myNodes[i];
 	    myNodes[i] = "";
 	    int n = internalNodes.length();
 	    for(int j = 0; j < n; j++)
 	     {
 	      String c = "" + internalNodes.charAt(j);
 	      if (s.contains(c))
 	    	 myNodes[i] = myNodes[i] + c;
 	     }
 	   if (print1)
 	    System.out.println(subtop[i] + " " + myNodes[i] );
 	   }	
    } 
   return myNodes;
    }
   
 
 //Subtopologies are compared to decide whether
 //or not they have the same set of leaves
 //and the same set of inner nodes.

public static void comparison(String[] Leaves1, 
		                       String[] Leaves2,   
		                       String[] Inner1, 
		                       String[] Inner2) 
 {
  System.out.println("\nComparison of leaves and nodes");
  //Print everything
  Boolean printAll = false;
  int m = Leaves1.length;
  int n = Leaves2.length;
  for(int i = 0; i < m; i++)
  {
   for(int j = 0; j < n; j++) 
   {
	if ( Leaves1[i].equals(Leaves2[j])) 
	{
	 System.out.println(
	 "Sets of leaves are EQUAL " + i + 
	 "  " + Leaves1[i] + "  " + j + "  " + Leaves2[j] );
	 if ( Inner1[i].equals(Inner2[j]))   
			System.out.println(
		        "Sets of inner nodes are EQUAL " + i +
	    "  " + Inner1[i] + "  " + j + "  " + Inner2[j] );
	  else 
			System.out.println(
		        "Sets of inner nodes are DIFFERENT " + i +
		"  " + Inner1[i] + "  " + j + "  " + Inner2[j] );	 
	}
	else 
		if (printAll)
		System.out.println(
	        "Sets of leaves are DIFFERENT " + i +
	   "  " + Leaves1[i] + "  " + j + "  " + Leaves2[j] );	
   }
  }	 
 }
 
 
 public static void main(String[] args) 
 {   
   //A topology is generated 
   String topo1 = generateTopology();
   System.out.println(topo1 + " -> " 
	          + simplify(topo1));
   //Length of String   topology
   m1 = topo1.length();
   //Subtopologies are found
   String[] subtopologies1 = new String[m1];
   subtopologies1 = subtopologies(topo1);
   m1 = counterSubtop;
   /*
   String[] Nodes1 = new String[m1];
   innerNodes(subtopologies1); */

   System.out.println();
   String[] Leaves1 = new String[m1];
   Leaves1 = leaves(m1,subtopologies1);
   String[] Inner1 = new String[m1];
   Inner1 = inner(m1,subtopologies1);
   //Chars of Leaves are rewritten in alphabet order
   System.out.println("\nSubtopologies, Leaves, " +
   		             "Inner nodes" );
   for(int i = 0; i < m1; i++)
   {
	 if (subtopologies1[i] == null);
	  else
	  {   
	   	  
	   System.out.println(i + "   " + subtopologies1[i]
			   + "   " +  Leaves1[i] + "   " + Inner1[i]);
	  }
   }
   
   System.out.println();
  
   //A second topology is generated

   String topo2 = generateTopology();
   System.out.println(topo2 + " -> " 
	          + simplify(topo2));
   //Length of String   topology
   m2 = topo2.length();
   //Subtopologies are found
   String[] subtopologies2 = new String[m2];
   subtopologies2 = subtopologies(topo2);
   m2 = counterSubtop;
   System.out.println();
   String[] Leaves2 = new String[m2];
   Leaves2 = leaves(m2,subtopologies2);
   String[] Inner2 = new String[m2];
   Inner2 = inner(m2,subtopologies2);
   //Chars of Leaves are rewritten in alphabet order
   System.out.println("\nSubtopologies, Leaves, " +
   		             "Inner nodes" );
   for(int i = 0; i < m2; i++)
   {
	 if (subtopologies1[i] == null);
	  else
	  {   
	   	  
	   System.out.println(i + "   " + subtopologies2[i]
			   + "   " +  Leaves2[i] + "   " + Inner2[i]);
	  }
   }   
  comparison(Leaves1,Leaves2, Inner1,Inner2); 
 }
}//End of main class K31  SubTopol4
