
//Program K16 TopoMutation.
//This program generates a random binary tree topology  
//and produces a mutation, whose output must be
//a binary tree topology with the same set of leaves.

//A mutation results from deleting 
//a given number of leaves
//to reinsert them anew in 
//other places chosen at random.


import java.util.Random;


public class TopoMutation
{

 //Link's notation for topologies
 private static   String   parent = 
		 "((Aa,Ba),Ca)";
 private static  String leaves =  
		 "ABCDEFGHIJKLMNOPRSTUVWXYZ";
 private static String internalNodes = 
		 "abcdefghijklmnopqrstuvwxyz";
 //Next leaf to be grafted	
 private static  int nextLeafIndex = 2;
 //Next internal node to be inserted
 private static  int nextIntNodeIndex = 0;
 
 private static  Random r = new Random();
 
 private static boolean print1 = false;
 private static boolean print2 = false; 
 
 //This is the original topology:
 private static  String original = "";
 
 //============Size of topology ================
 //Number of leaves to be inserted to parent
 private static  int nAdditionalLeaves = 10; 
 //=========================================
 
 //Number of leaves to be deleted and reinserted
 //to create a mutation
 private static int nUndoRedo = 0;
 
 
 //Detect internal nodes, different than leaves
 private static boolean charIsInternalNode(char c)
 {
  String s = "" + c;	 
  if (internalNodes.contains(s)) return true;
  else return false;
 }
 
//Simplification from Link notation to Newick's notation:
 // (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)
private static String simplify(String newTopo)
{
 String topo = "";
//length of string newTopo
int l = newTopo.length();
 for(int i = 0; i < l; i++ )
  {
   //newTopo is scanned char by char	 
  char c = newTopo.charAt(i);
  //Is c an internal node to be deleted?
  if (charIsInternalNode(c));
   else //add char c to output
    topo = topo + c;
  }
 return topo;	 
}
 
//Rule 2 is used to generate a new topology:
//(S)I -> ((S)N,FN)I.
//S = subtopology, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string parentTopo
private static String useRule2(String parentTopo, int i, 
        char F, char N)
{
 String newTopo = "";
 char I = parentTopo.charAt(i);
 if (print1)
 System.out.println("i = " + i + " I = " + I);
 //Subtopology S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = i;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
   String a = "" + parentTopo.charAt(position);
   
   if (print1)
   System.out.println("a = " + a + " p = " + pr);
   if (a.equals(pr) ) 
   {
	counter++;
	S = a + S;
   }
   else go = false;
  }//End of while
 if (print1)
 System.out.println("counter = " + counter);
 //The complete subtopology is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String a = "" + parentTopo.charAt(position);
   if (a.equals(pr))   counter++;
   if (a.equals(pl))  counter--;
   S = a + S;
   if (print1)
   System.out.println("Subtopo = " + S);
   position = position -1;
   }
 //Replacement rule 2:
 //((S)n,Fn)i -> (S)i
//where   F are leaves, n and i are  inner nodes,
//(S) is a complex structure. 
 String SnFni	 =   S  + I; 
 String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
 //Rule 1 is executed
 newTopo = parentTopo.replace(SnFni, replacement);		
 //If I links to a complex structure
 return newTopo;	 
}

//From parentTopo, a topology in Newick's notation,
//a Child  topology is generated by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String parentTopo) 
{
	
 nextLeafIndex++;
 char nextLeaf= leaves.charAt(nextLeafIndex);
 //Next internal node to be inserted
 nextIntNodeIndex++;
 char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
	
	
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new topology generated by replacement
 String newTopo = "";
 //length of parent topology
 int l = parentTopo.length();
 //counter of links = 0;
 int counter = 0;
 //Links are counted
 for(int i = 0; i < l; i++ )
  {
   //parentTopo is scanned char by char	 
   char I = parentTopo.charAt(i);
	//Is I an internal node that denotes a link?
	if (charIsInternalNode(I)) counter++;	
  }
 //A link is chosen 
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
  {
    //parentTopo is scanned char by char	 
	char I = parentTopo.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = parentTopo.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTopo = useRule2(parentTopo, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTopo = parentTopo.replace(link, replacement);		
		 }
		if (print2)
		System.out.println(newTopo + " -> " 
		                       + simplify(newTopo));			
 
	  
		 i = l;
	   }
	  counter++;	
	} 
  } 
 return newTopo;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply the antirule1:
//(An,Fn)i  -> Ai
private static String  antiRule1(String topo, 
		int positionLeaf, int positionNode)
{
 if (print2)
  System.out.println("AntiRule1");
 int beginIndex = positionLeaf - 4;
 int endIndex = positionNode + 3;
 String AnFni = topo.substring(beginIndex, endIndex);
 if (print2)
  System.out.println("Anfni = " + AnFni);
 String Ai = "" + topo.charAt(beginIndex+1)
                + topo.charAt(endIndex-1);
 if (print2)
  System.out.println("Ai = " + Ai);
 topo = topo.replace(AnFni, Ai);		
	return topo;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply the antirule2:
//((S)n,Fn)i -> (S)i
//positionNode refers to the last appearing of node n
private static String  antiRule2(String topo, 
		int positionLeaf, int positionNode)
{
 if (print2)
	System.out.println("AntiRule2");
 //Leaf to be deleted	
 String leaf = "" + leaves.charAt(nextLeafIndex);
 if (print2)
  System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
	            + internalNodes.charAt(nextIntNodeIndex);	
 int firstPosition = topo.indexOf(innerNode);
 String newTopo = "";
 char i = topo.charAt(positionNode+2);
 if (print2)
 System.out.println("FirstPosition = " + 
                      firstPosition + " i = " + i);
 //Subtopology S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = firstPosition;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
    String as = "" + topo.charAt(position);
   
   if (print2)
   System.out.println("a = " + as + " p = " + pr);
   if (as.equals(pr) ) 
   {
	counter++;
	S = as + S;
   }
   else go = false;
  }//End of while
 if (print2)
 System.out.println("counter = " + counter);
 //The complete subtopology is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String as = "" + topo.charAt(position);
   if (as.equals(pr))   counter++;
   if (as.equals(pl))  counter--;
   S = as + S;
   if (print2)
   System.out.println("Subtopo = " + S);
   position = position -1;
   }
  String n = innerNode;
 //Replacement antiRule 2:
 //((S)n,Fn)i -> (S)i
 //where A and F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =  "("+ S  + n + "," + leaf + n + ")" + i ; 
 String	Si = S + i;
 if (print2)
  System.out.println("\nSnFni = " + SnFni + "\n Si = " + Si);
 //Rule 1 is executed
 newTopo = topo.replace(SnFni, Si);		
 return newTopo;
}



//The last inserted  leaf is deleted from topology.
private static String deleteLeaf(String topology) 
{
  //Leaf to be deleted	
  String leaf = "" + leaves.charAt(nextLeafIndex);
  if (print2)
  System.out.println("\nLeaf to be deleted = " + leaf);
  //Accompanying inner node  
  String innerNode = "" 
            + internalNodes.charAt(nextIntNodeIndex);
  if (print2)
  System.out.println("Accompanying inner  node = " 
                    + innerNode);		
  //new topology = output
  String newtopology = "";
  //length of  topology as string
  int l = topology.length();
  //Last positions are found
  int positionLeaf = -1;
  int positionNode = -1;  
  //topology is scanned char by char
  for(int i = 0; i < l; i++ )
   {
	 String c = "" + topology.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
   }
  if (print2)
   {
    System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
    System.out.println("Position of " + innerNode + 
          " = " + positionNode);
   }
  int firstPosition = topology.indexOf(innerNode);
  //Check for the char before firstPosition	
  char a = topology.charAt(firstPosition - 1);
  //If I links to a complex structure
  if (a == ')')
  newtopology = antiRule2(topology,positionLeaf,positionNode);
  else  //Apply antiRule1 for simple cases
	  newtopology = antiRule1(topology,positionLeaf,positionNode);
  nextLeafIndex--;
  nextIntNodeIndex--;
  return newtopology;	
}


/**
 * @param args
 */
public static void main(String[] args) 
{
 //A topology is generated	
 if (print2)
  System.out.println("Parent topology = \n" + parent
		 + " -> "    + simplify(parent)); 
 String topology = parent;

 for(int i = 0; i < nAdditionalLeaves; i++ )	 
   topology = generateChild(topology);
 //This original topology is next subject to mutation
 original = topology;
 if (print2)
  System.out.println("\nOriginal   topology = \n" 
		                + original  );
 //Number of leaves to be deleted and then reinserted
 if (nAdditionalLeaves == 1) nUndoRedo = 1;
  else
   nUndoRedo = r.nextInt(nAdditionalLeaves-1) + 1;
 //nUndoRedo = 13;
 
 //Mutation has two stages:
 //deletion of some leaves that undoes generating work + 
 //re-insertion in places chosen at random.
 
 if (print2)
  System.out.println("Number of leaves to" +
  		" be deleted and then reinserted = " +
  		nUndoRedo); 
 //Some leaves are deleted:
 if (print2)
 System.out.println("Some leaves are deleted");
 for(int i = 0; i < nUndoRedo; i++ )	 
 {
  //Last inserted leaf is deleted.
  topology = deleteLeaf(topology);
  if (print2)
   System.out.println(topology);
 }
//Some leaves are reinserted
 if (print2)
  System.out.println("Deleted leaves are reinserted");
 for(int i = 0; i < nUndoRedo; i++ )	 
  {
   topology = generateChild(topology);
  }
 //Original and mutant topologies are compared.
 String mutant = topology;
 System.out.println("\nOriginal and mutant topologies = \n" 
  + original + "\n" + mutant);
 if (original.equals(mutant))
	  System.out.println(
			  "Original and mutant topologies coincide");
 else
	  System.out.println(
			  "\nOriginal and mutant topologies are different");    
}

}//End of main class K16 TopoMutation