
//Program K51 TopoMutation3.
//This program generates a random binary tree topology  
//and produces a mutation, whose output must be
//a binary tree topology over the same set of leaves.

//At the beginning it is decided with which order to work:
//normal, reversed or permuted.
//Next, a tree is synthesized
//and a mutation is made.
//
//A mutation results from deleting 
//a given number of leaves
//to reinsert them anew in 
//other places chosen at random.
//Leaves can be deleted in reversed order of insertion
//or at random.

//Seed tree, with three leaves, is conserved.
//Example in ordinary order ((Aa,Ba),Ca).



import java.util.Random;


public class TopoMutation3
{

 //Link's notation for topologies
 private static   String   seed = 
		 "((Aa,Ba),Ca)";
 private static  String leaves =  
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 private static String internalNodes = 
		 "abcdefghijklmnopqrstuvwxyz";
 //Next leaf to be grafted	
 private static  int nextLeafIndex = 2;
 //Next internal node to be inserted
 private static  int nextIntNodeIndex = 0;
 
 
 
 //Number of leaves to be inserted to seed:
 //to get a tree with 4 leaves, add one leaf to seed
 private static  int nAdditionalLeaves = 3;
 

 //Number of leaves to be deleted and reinserted
 //to create a mutation.
 //mustbe less than nAdditionalLeaves
 private static int nUndoRedo = 0;
 
 
 //Leaves that remain in the topology
 private static  String presentLeaves =   "ABC";
 private static String presentNodes =  "a";
 //Founder leaves ABC or equivalent cannot be removed
 private static String deletableLeaves = "";
 //Deleted leaves are then reinserted
 private static String reinsertingLeaves = "";
 private static String reinsertingNodes = "";
 
 private static  Random r = new Random();
 
 private static boolean print1 = false;
 private static boolean antiRule4 = false; 
 
 //This is the original topology:
 private static  String original = "";
 

 

//String s ( a char) is deleted from mother
public static String delete(String s, String mother)
{
String output = "";
int l = mother.indexOf(s);
output = mother.substring(0, l)  
		 + mother.substring(l+1) ;
return output;
}



//String leaves is reversed
public static String reverse(String leaves)
{
String reversed = "";
int l = leaves.length();
for(int i = l-1; i >= 0; i--)
 {
	String c = "" + leaves.charAt(i); 
	reversed = reversed + c;
 }
if (print1)
System.out.println("Reverse order = " + reversed);
seed =  "((Za,Ya),Xa)";
if (print1)
System.out.println("seed = " + seed);
presentLeaves =   "ZYX";
return reversed;
}


//String leaves is permuted
public static String permuteOrder(String leaves)
{
	 if (print1)
System.out.println("original and permuted order = \n" 
                         + leaves);
String permuted = "";
int l = leaves.length();
int m = l;
for(int i = 0; i < l; i++)
 {
	int d = r.nextInt(m) ;
	String c = "" + leaves.charAt(d); 
	leaves = delete(c,leaves);
	m = m-1;
	permuted = permuted + c;
 }
if (print1)
System.out.println(permuted);
String first = "" + permuted.charAt(0);
String second = "" + permuted.charAt(1);
String third = "" + permuted.charAt(2); 
seed = "((" +first + "a," +second + "a)," +third + "a)";
if (print1)
System.out.println("seed = " + seed);
presentLeaves =   first + second + third;
return permuted;	
}


//Detect internal nodes, different than leaves
private static boolean charIsInternalNode(char c)
{
 String s = "" + c;	 
 if (internalNodes.contains(s)) return true;
 else return false;
}

//Simplification from Link notation to Newick's notation:
// (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)
private static String simplify(String newTree)
{
String tree = "";
//length of string newTree
int l = newTree.length();
for(int i = 0; i < l; i++ )
 {
  //newTree is scanned char by char	 
 char c = newTree.charAt(i);
 //Is c an internal node to be deleted?
 if (charIsInternalNode(c));
  else //add char c to output
   tree = tree + c;
 }
return tree;	 
}

//Rule 2 is used to generate a new tree:
//(S)I -> ((S)N,FN)I.
//S = subtree, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string seedTree
private static String useRule2(String seedTree, int i, 
       char F, char N)
{
String newTree = "";
char I = seedTree.charAt(i);
if (print1)
System.out.println("i = " + i + " I = " + I);
//Subtree S of link (S)I must be read 
String S = "";
//Counter of left seedhesis
int counter = 0;
String pl = "" + '(';
String pr = "" + ')';
//Position to be scanned
int position = i;
boolean go = true;
//All right seedheses to the near left of I are captured
while (go)
 {
  position = position -1;
  String a = "" + seedTree.charAt(position);
  
  if (print1)
  System.out.println("a = " + a + " p = " + pr);
  if (a.equals(pr) ) 
  {
	counter++;
	S = a + S;
  }
  else go = false;
 }//End of while
if (print1)
System.out.println("counter = " + counter);
//The complete subtree is captured:
//seedheses must balanced
while (counter > 0)
 {   
  String a = "" + seedTree.charAt(position);
  if (a.equals(pr))   counter++;
  if (a.equals(pl))  counter--;
  S = a + S;
  if (print1)
  System.out.println("Subtree = " + S);
  position = position -1;
  }
//Replacement rule 2:
//((S)n,Fn)i -> (S)i
//where   F are leaves, n and i are  inner nodes,
//(S) is a complex structure. 
String SnFni	 =   S  + I; 
String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
//Rule 1 is executed
newTree = seedTree.replace(SnFni, replacement);		
//If I links to a complex structure
return newTree;	 
}

//Input: seedTree, a tree in Newick's notation.
//Output : newTree =  seed tree + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String seedTree) 
{
	
nextLeafIndex++;
char nextLeaf = leaves.charAt(nextLeafIndex);
presentLeaves = presentLeaves + nextLeaf;
//Next internal node to be inserted
nextIntNodeIndex++;
char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
presentNodes = presentNodes + nextIntNode;
	
//link to be replaced
String link = "";
//String link is replaced by what:
String replacement = "";
//new tree generated by replacement
String newTree = "";
//length of seed tree
int l = seedTree.length();
//counter of links = 0;
int counter = 0;
//Links are counted
for(int i = 0; i < l; i++ )
 {
  //seedTree is scanned char by char	 
  char I = seedTree.charAt(i);
	//Is I an internal node that denotes a link?
	if (charIsInternalNode(I)) counter++;	
 }
//A link is chosen 
int k = r.nextInt(counter);
//System.out.println("Random number = " + k);
//new branch is grafted at k
counter = 0;  
for(int i = 0; i < l; i++ )
 {
   //seedTree is scanned char by char	 
	char I = seedTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = seedTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(seedTree, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTree = seedTree.replace(link, replacement);		
		 }
		 if (print1)
		System.out.println(newTree + " -> " 
		                       + simplify(newTree));			

	  
		 i = l;
	   }
	  counter++;	
	} 
 } 
return newTree;
}


//Input: seedTree, a tree in Newick's notation.
//Output : newTree =  seed tree + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String reinsertLeaf(String seedTree) 
{
nextLeafIndex++;
char nextLeaf = reinsertingLeaves.charAt(nextLeafIndex);
presentLeaves = presentLeaves + nextLeaf;
//Next internal node to be inserted
nextIntNodeIndex++;
char nextIntNode = reinsertingNodes.charAt(nextIntNodeIndex);
presentNodes = presentNodes + nextIntNode;
	
//link to be replaced
String link = "";
//String link is replaced by what:
String replacement = "";
//new tree generated by replacement
String newTree = "";
//length of seed tree
int l = seedTree.length();
//counter of links = 0;
int counter = 0;
//Links are counted
for(int i = 0; i < l; i++ )
 {
  //seedTree is scanned char by char	 
  char I = seedTree.charAt(i);
  //Is I an internal node that denotes a link?
  if (charIsInternalNode(I)) counter++;	
 }
//A link is chosen 
int k = r.nextInt(counter);
//System.out.println("Random number = " + k);
//new branch is grafted at k
counter = 0;  
for(int i = 0; i < l; i++ )
 {
   //seedTree is scanned char by char	 
	char I = seedTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = seedTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(seedTree, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTree = seedTree.replace(link, replacement);		
		 }
		 if (print1)
		System.out.println(newTree + " -> " 
		                       + simplify(newTree));			

	  
		 i = l;
	   }
	  counter++;	
	} 
 } 
return newTree;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply the antirule1:
//(An,Fn)i  -> Ai
private static String  antiRule1(String tree, 
		int positionLeaf, int positionNode)
{
	 if (print1)
System.out.println("AntiRule1");
int beginIndex = positionLeaf - 4;
int endIndex = positionNode + 3;
String AnFni = tree.substring(beginIndex, endIndex);
if (print1)
System.out.println("Anfni = " + AnFni);
String Ai = "" + tree.charAt(beginIndex+1)
               + tree.charAt(endIndex-1);
if (print1)
System.out.println("Ai = " + Ai);
tree = tree.replace(AnFni, Ai);
if (print1)
System.out.println("Output tree = \n" + tree);
return tree;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply the antirule2:
//((S)n,Fn)i -> (S)i
//positionNode refers to the last appearing of node n
private static String  antiRule2(String tree, 
		int positionLeaf, int positionNode)
{
	
	 if (print1)
		 System.out.println("AntiRule2");
//Leaf to be deleted	
String leaf = "" + tree.charAt(positionLeaf);
if (print1)
System.out.println("\nDeletion = " + leaf);
//Accompanying inner node  
String innerNode = "" 
	            + tree.charAt(positionNode);	
int firstPosition = tree.indexOf(innerNode);
String newTree = "";
String i = ""; //BUG?????
if (positionNode +2 < tree.length())
        i = ""+ tree.charAt(positionNode+2);
if (print1)
System.out.println("FirstPosition = " + 
                     firstPosition + " i = " + i);
//Subtree S of link (S)I must be read 
String S = "";
//Counter of left seedhesis
int counter = 0;
String pl = "" + '(';
String pr = "" + ')';
//Position to be scanned
int position = firstPosition;
boolean go = true;
//All right seedheses to the near left of I are captured
while (go)
 {
  position = position -1;
   String as = "" + tree.charAt(position);
  
  if (print1)
  System.out.println("a = " + as + " p = " + pr);
  if (as.equals(pr) ) 
  {
	counter++;
	S = as + S;
  }
  else go = false;
 }//End of while
if (print1)
System.out.println("counter = " + counter);
//The complete subtree is captured:
//seedheses must balanced
while (counter > 0)
 {   
  String as = "" + tree.charAt(position);
  if (as.equals(pr))   counter++;
  if (as.equals(pl))  counter--;
  S = as + S;
  if (print1)
  System.out.println("Subtree = " + S);
  position = position -1;
  }
 String n = innerNode;
//Replacement antiRule 2:
//((S)n,Fn)i -> (S)i
//where A and F are leaves, n and i are  inner nodes,
//(S) is a complex structure. 
String SnFni	 =  "("+ S  + n + "," + leaf + n + ")" + i ; 
String	Si = S + i;
if (print1)
 System.out.println("\nSnFni = " + SnFni + "\n Si = " + Si);
//Rule 1 is executed
newTree = tree.replace(SnFni, Si);	
if (print1)
System.out.println("Output tree = \n" + newTree);
return newTree;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply   antirule3:
//(Fn,An)i  -> Ai
private static String  antiRule3(String topo, 
		int positionLeaf, int positionNode)
{
	 if (print1)
System.out.println("AntiRule3"
		+ "\nInput topology = \n" + topo);
int beginIndex = positionLeaf - 1;
int endIndex = positionNode + 6;
String FnAni = topo.substring(beginIndex, endIndex);
if (print1)
System.out.println("FnAni = " + FnAni);
String Ai = "" + topo.charAt(beginIndex+4)
         + topo.charAt(endIndex-1);
if (print1)
System.out.println("Ai = " + Ai);
topo = topo.replace(FnAni, Ai);	
if (print1)
System.out.println("Output topology = \n" + topo);
return topo;
}




//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply antirule4:
//(Fn,(S)n)i -> (S)i
private static String  antiRule4(String tree, 
		int positionLeaf, int positionNode)
{

//Leaf to be deleted	
String leaf = "" + tree.charAt(positionLeaf);

//Accompanying inner node  
String innerNode = "" 
		            + tree.charAt(positionNode);
if (print1)
System.out.println("\nAntiRule4"
		+ "\nLeaf to be deleted = " + leaf
		 + " at position " + positionLeaf
		+  "\nAccompanying node = " + innerNode
			 + " at first position " + positionNode);
//Second position of accompanying node
int positionNode2 = 0;
int l = tree.length();
boolean go = true;
//tree is scanned char by char
for(int i = positionNode + 1; i < l; i++ )
{
	 String c = "" + tree.charAt(i);
	
	 if (go)
	 {
	  if (c.equals(innerNode))
		 {
		  positionNode2 = i;
	
       go = false;
       if (print1)
     		 System.out.println(
     			"Second position of accompanying node = " + 
     			 	 positionNode2 + " i = " + i);
		 }
	 }
	
}

	
int beginIndex = positionLeaf - 1;
int endIndex = positionNode2+2;
String i = ""; 
if (endIndex < tree.length())
	         i = ""+ tree.charAt(endIndex);
	 
if (print1)
System.out.println("EndIndex = " + 
	                      endIndex + " i = " + i);
	 
String FnSni	 = tree.substring(beginIndex, endIndex) + i;
if (print1)
System.out.println("FnSni = " + FnSni);
String S = tree.substring(beginIndex+4, endIndex-2);
if (print1)
	System.out.println("S = " + S);
String	Si = S + i;
if (print1)
	 System.out.println("Si = " + Si);
	 
String newTree = tree.replace(FnSni, Si);
if (print1)
	 System.out.println("Output tree = \n" + newTree); 
//System.exit(0);	 
return newTree;
}



//The last inserted  leaf is deleted from tree.
private static String deleteLastLeaf(String tree) 
{
//Leaf to be deleted	
String leaf = "" + leaves.charAt(nextLeafIndex);
if (print1)
System.out.println("\nLeaf to be deleted = " + leaf);
//Accompanying inner node  
String innerNode = "" 
       + internalNodes.charAt(nextIntNodeIndex);
if (print1)
System.out.println("Accompanying inner  node = " 
               + innerNode);		
//new tree = output
String newTree = "";
//length of  tree as string
int l = tree.length();
//Last positions are found
int positionLeaf = -1;
int positionNode = -1;  
//tree is scanned char by char
for(int i = 0; i < l; i++ )
  {
	 String c = "" + tree.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
  }
if (print1)
System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
if (print1)
System.out.println("Position of " + innerNode + 
     " = " + positionNode);
int firstPosition = tree.indexOf(innerNode);
//Check for the char before firstPosition	
char a = tree.charAt(firstPosition - 1);
//If I links to a complex structure
if (a == ')')
  newTree = antiRule2(tree,positionLeaf,positionNode);
 else  //Apply antiRule1 for simple cases
	  newTree = antiRule1(tree,positionLeaf,positionNode);
nextLeafIndex--;
nextIntNodeIndex--;
return newTree;	
}



//A random leaf is deleted from tree. 
private static String deleteRandomLeaf(String tree) 
{
	 if (print1)
	 {
System.out.println("\nMethod deleteRandomLeaf"
		 + "\ntree = " + tree
		+ "\ndeletableLeaves = " + deletableLeaves);
	 }
int l = deletableLeaves.length();
int position = r.nextInt(l);
//Leaf to be deleted	
String leaf = "" + deletableLeaves.charAt(position);
int nodeIndex = tree.indexOf(leaf) + 1;
//Accompanying inner node  
String innerNode = "" 
       + tree.charAt(nodeIndex);

deletableLeaves = delete(leaf,deletableLeaves);
presentLeaves = delete(leaf,presentLeaves);
reinsertingLeaves = reinsertingLeaves + leaf;
reinsertingNodes = reinsertingNodes + innerNode;


if (print1)
System.out.println("\nLeaf to be deleted = " + leaf
		+ "\nReinserting leaves = " + reinsertingLeaves
    + "\nAccompanying inner  node = "   + innerNode
    + "\nReinserting nodes = " + reinsertingNodes);		
//new tree = output
String newTree = "";
//length of  tree as string
l = tree.length();
//Last positions are found
int positionLeaf = tree.indexOf(leaf);
int positionNode = positionLeaf + 1;  


if (print1)
System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
if (print1)
System.out.println("Position of " + innerNode + 
     " = " + positionNode);
//Which set of deleting rules to apply:
//1 and 2 else 3 and 4?
String previousChar = "" + tree.charAt(positionLeaf-1);
if (previousChar.equals(",")) //Apply antiRules 1 and 2
{
	if (print1)
		System.out.println("Apply antiRules 1 and 2");
int firstPosition = tree.indexOf(innerNode);
//Check for the char before firstPosition	
char a = tree.charAt(firstPosition - 1);
//If I links to a complex structure
if (a == ')')
newTree = antiRule2(tree,positionLeaf,positionNode);
else  //Apply antiRule1 for simple cases
	  newTree = antiRule1(tree,positionLeaf,positionNode);
nextLeafIndex--;
nextIntNodeIndex--;
}
else //Apply antiRules 3 and 4
{
	if (print1)
		System.out.println("Apply antiRules 3 and 4");
int firstPosition = tree.indexOf(innerNode);
//Check for the char before firstPosition	
char a = tree.charAt(firstPosition + 2);
//If I links to a complex structure
if (a == '(')
	 newTree = antiRule4(tree,positionLeaf,positionNode);
	    else  //Apply antiRule3 for simple cases
newTree = antiRule3(tree,positionLeaf,positionNode);
nextLeafIndex--;
nextIntNodeIndex--;	 
}

if (print1)
System.out.println("Tree after deletion" + newTree);
return newTree;	
}




//Mutation = 
//Random deletions are reinserted.
public static String  mutationRandomDeletion(
		           String tree, int nAdditionalLeaves)
{
//Some leaves are deleted: random order
	 if (print1)
System.out.println("Some leaves are deleted AT RANDOM");
for(int i = 0; i < nUndoRedo; i++ )	 
		 {
		  //Last inserted leaf is deleted.
		  tree = deleteRandomLeaf(tree);
		 }
//Deleted leaves are reinserted
nextLeafIndex = -1;
nextIntNodeIndex = -1;
if (print1)
System.out.println("Deleted leaves are reinserted");
for(int i = 0; i < nUndoRedo; i++ )	 
	   {
		  
	     tree = reinsertLeaf(tree);
	   }
return tree;
}



//Mutation:
//some leaves are deleted and then reinserted 
//in random order.
//Deleting can happen in either of two ways:
//Reversing the order of synthesis else
//in random order.
public static void mutation(String topology, 
		                       int nAdditionalLeaves)
{
 //seedal 3 leaves are not removed	
 deletableLeaves = presentLeaves.substring(3);
 //Number of leaves to be deleted and then reinserted
 if (nAdditionalLeaves == 1) nUndoRedo = 1;
  else
   nUndoRedo = r.nextInt(nAdditionalLeaves-1) + 1;
 System.out.println("\nNumber of leaves to delete =  "
                     + nUndoRedo);
 
 //choose: random order else reverse to that of synthesis
 double pReverseSynth = 0;
 double s = r.nextDouble();
 if ( s < pReverseSynth)//Choose deletion in reversed order
  {
   //Some leaves are deleted: 
   //reversed order to that of synthesis.
   System.out.println("Leaves are deleted " +
   		                        "in reverse order");
    for(int i = 0; i < nUndoRedo; i++ )	 
	 {
	  //Last inserted leaf is deleted.
	  topology = deleteLastLeaf(topology);
	  System.out.println(topology);
	 }
   //deleted leaves are reinserted
   System.out.println("\nDeleted leaves are reinserted");
   for(int i = 0; i < nUndoRedo; i++ )	 
    {
	  topology = generateChild(topology);
    } 
  }
 else
  topology = mutationRandomDeletion(topology, nUndoRedo);
 //Output
 String mutant = topology;
 System.out.println("\nOriginal and mutant topologies = \n" 
		  + original + "\n" + mutant);
 if (original.equals(mutant))
			  System.out.println("Mutation made nothing");
	   else
			  System.out.println("\nMutation worked");	
}


public static void main(String[] args) 
{
 //Are you going to work with the given order,
 //a reverse one, or with a permutation?
 double pGivenOrder = 0;
 double pReverseOrder = 0;
 double a = pGivenOrder + pReverseOrder;
 //Implicit: probability of permutation =
 //1 - (pGivenOrder+pReverseOrder).
 //Election is made
 double s = r.nextDouble();
 if ( s < pGivenOrder); //Do nothing = order is not changed
  else if ( s < a )	   leaves = reverse(leaves);
	 else leaves = permuteOrder(leaves);//permute if  s > a	 
 //Change of format
 System.out.println("seed topology = \n" + seed
		 + " -> "    + simplify(seed)); 		  	
 //A topology topology is generated	

 String topology = seed;

 for(int i = 0; i < nAdditionalLeaves; i++ )	 
   topology = generateChild(topology);
 //This original topology topology is next subject to mutation
 original = topology;
 mutation(topology, nAdditionalLeaves);
 if (antiRule4)
	 System.out.println("ANTIRULE 4");
}
}//End of main class K51 TopoMutation3