
//Program K22 TopoMutation7.
//This program generates a random binary tree topology,
//adds lengths to links to get a tree,
//and produces a mutation, whose output must be
//a well formed binary tree  over the same set of leaves.

//At the beginning it is decided 
//with which order of the alphabet of leaves to work:
//normal, reversed or permuted.
//Next, a tree is synthesized
//and a mutation is made.
//
//A mutant topology results from deleting 
//a given number of leaves
//to reinsert them anew in 
//other places chosen at random.
//Leaves can be deleted in reversed order of insertion
//or at random.

//Lengths of links are also changed.

//Parent tree, with three leaves, is conserved.
//Example in ordinary order ((Aa,Ba),Ca).





import java.util.Random;


public class TopoMutation7
{

 //Link's notation for topologies
 private static   String   parent = 
		 "((Aa,Ba),Ca)";
 private static  String leaves =  
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 private static String internalNodes = 
		 "abcdefghijklmnopqrstuvwxyz";
 //Next leaf to be grafted	
 private static  int nextLeafIndex = 2;
 //Next internal node to be inserted
 private static  int nextIntNodeIndex = 0;
 
 //Leaves that remain in the topology
 private static  String presentLeaves =   "ABC";
 private static String presentNodes =  "a";
 //Founder leaves ABC or equivalent cannot be removed
 private static String deletableLeaves = "";
 
//Mutation rate per site (= entry) per generation
static private double    m = 0.5;

 private static  Random r = new Random();
 //The control parameter of the random generator for int
 private static int maxLength = 10;
 
 private static boolean print1 = false;
 private static boolean print2 = false; 
 
 
 //This is the original topology:
 private static  String original = "";
 
 //===========Size of the tree=========== 
 //Number of leaves to be inserted
 private static  int nLeavesToBeInserted = 3; 
 //=====================================
 
 
 
 //Number of leaves to be deleted and reinserted
 private static int nLeavesToDelete = 3;
 
//Number of leaves that must go in all trees > 3	
private static	int numberOfLeaves = 6;


 //Length of the link that glues a leaf to the tree
 private static int lengthsLeaves[]  = 
		                new int[10];
 //Length of the link that glues 
 //an inner node to another one
 private static int lengthsInnerLinks[]  = 
		               new int[10];

//String s ( a char) is deleted from mother
public static String delete(String s, String mother)
{
 String output = "";
 int l = mother.indexOf(s);
 output = mother.substring(0, l)  
		 + mother.substring(l+1) ;
 return output;
}



//String leaves is reversed
public static String reverse(String leaves)
{
 String reversed = "";
 int l = leaves.length();
 for(int i = l-1; i >= 0; i--)
  {
	String c = "" + leaves.charAt(i); 
	reversed = reversed + c;
  }
 if (print1)
  System.out.println("Reverse order = " + reversed);
 parent =  "((Za,Ya),Xa)";
 if (print1)
	System.out.println("Parent = " + parent);
 presentLeaves =   "ZYX";
 return reversed;
}


//String leaves is permuted
public static String permuteOrder(String leaves)
{
	if (print1)
	System.out.println("original and permuted order = \n" 
                          + leaves);
 String permuted = "";
 int l = leaves.length();
 int m = l;
 for(int i = 0; i < l; i++)
  {
	int d = r.nextInt(m) ;
	String c = "" + leaves.charAt(d); 
	leaves = delete(c,leaves);
	m = m-1;
	permuted = permuted + c;
  }
 if (print1)
  System.out.println(permuted);
 String first = "" + permuted.charAt(0);
 String second = "" + permuted.charAt(1);
 String third = "" + permuted.charAt(2); 
 parent = "((" +first + "a," +second + "a)," +third + "a)";
 if (print1)
  System.out.println("Parent = " + parent);
 presentLeaves =   first + second + third;
 return permuted;	
}


 //Detect internal nodes, different than leaves
 private static boolean charIsInternalNode(char c)
 {
  String s = "" + c;	 
  if (internalNodes.contains(s)) return true;
  else return false;
 }
 
//Simplification from Link notation to Newick's notation:
 // (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)
private static String simplify(String newTree)
{
 String tree = "";
//length of string newTree
int l = newTree.length();
 for(int i = 0; i < l; i++ )
  {
   //newTree is scanned char by char	 
  char c = newTree.charAt(i);
  //Is c an internal node to be deleted?
  if (charIsInternalNode(c));
   else //add char c to output
    tree = tree + c;
  }
 return tree;	 
}
 
//Rule 2 is used to generate a new tree:
//(S)I -> ((S)N,FN)I.
//S = subtree, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string parentTree
private static String useRule2(String parentTree, int i, 
        char F, char N)
{
 String newTree = "";
 char I = parentTree.charAt(i);
 if (print1)
  System.out.println("i = " + i + " I = " + I);
 //Subtree S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = i;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
   String a = "" + parentTree.charAt(position);
   
   if (print1)
    System.out.println("a = " + a + " p = " + pr);
   if (a.equals(pr) ) 
   {
	counter++;
	S = a + S;
   }
   else go = false;
  }//End of while
 if (print1)
  System.out.println("counter = " + counter);
 //The complete subtree is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String a = "" + parentTree.charAt(position);
   if (a.equals(pr))   counter++;
   if (a.equals(pl))  counter--;
   S = a + S;
   if (print1)
    System.out.println("Subtree = " + S);
   position = position -1;
   }
 //Replacement rule 2:
 //((S)n,Fn)i -> (S)i
//where   F are leaves, n and i are  inner nodes,
//(S) is a complex structure. 
 String SnFni	 =   S  + I; 
 String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
 //Rule 1 is executed
 newTree = parentTree.replace(SnFni, replacement);		
 //If I links to a complex structure
 return newTree;	 
}

//Input: parentTopol, a topology in Newick's notation.
//Output : newTopol =  parent topology + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String parentTopol) 
{
	
 nextLeafIndex++;
 char nextLeaf = leaves.charAt(nextLeafIndex);
 presentLeaves = presentLeaves + nextLeaf;
 //Next internal node to be inserted
 nextIntNodeIndex++;
 char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
 presentNodes = presentNodes + nextIntNode;
	
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new Topol generated by replacement
 String newTopol = "";
 //length of parent Topol
 int l = parentTopol.length();
 //counter of links = 0;
 int counter = 0;
 //Links are counted
 for(int i = 0; i < l; i++ )
  {
   //parentTopol is scanned char by char	 
   char I = parentTopol.charAt(i);
	//Is I an internal node that denotes a link?
	if (charIsInternalNode(I)) counter++;	
  }
 //A link is chosen 
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
  {
    //parentTopol is scanned char by char	 
	char I = parentTopol.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = parentTopol.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTopol = useRule2(parentTopol, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTopol = parentTopol.replace(link, replacement);		
		 }
		if (print1)
		System.out.println(newTopol + " -> " 
		                       + simplify(newTopol));			
 
	  
		 i = l;
	   }
	  counter++;	
	} 
  } 
 return newTopol;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply the antirule1:
//(An,Fn)i  -> Ai
private static String  antiRule1(String tree, 
		int positionLeaf, int positionNode)
{
 if (print1)
  System.out.println("AntiRule1");
 int beginIndex = positionLeaf - 4;
 int endIndex = positionNode + 3;
 String AnFni = tree.substring(beginIndex, endIndex);
 if (print1)
  System.out.println("Anfni = " + AnFni);
 String Ai = "" + tree.charAt(beginIndex+1)
                + tree.charAt(endIndex-1);
 if (print1)
  System.out.println("Ai = " + Ai);
 tree = tree.replace(AnFni, Ai);
 if (print1)
  System.out.println("Output tree = \n" + tree);
 return tree;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply the antirule2:
//((S)n,Fn)i -> (S)i
//positionNode refers to the last appearing of node n
private static String  antiRule2(String tree, 
		int positionLeaf, int positionNode)
{
 if (print1)
	System.out.println("AntiRule2");
 //Leaf to be deleted	
 String leaf = "" + tree.charAt(positionLeaf);
 if (print1)
  System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
	            + tree.charAt(positionNode);	
 int firstPosition = tree.indexOf(innerNode);
 String newTree = "";
 String i = ""; //BUG?????
 if (positionNode +2 < tree.length())
         i = ""+ tree.charAt(positionNode+2);
 if (print2)
  System.out.println("FirstPosition = " + 
                      firstPosition + " i = " + i);
 //Subtree S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = firstPosition;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
    String as = "" + tree.charAt(position);
   
   if (print2)
   System.out.println("a = " + as + " p = " + pr);
   if (as.equals(pr) ) 
   {
	counter++;
	S = as + S;
   }
   else go = false;
  }//End of while
 if (print2)
  System.out.println("counter = " + counter);
 //The complete subtree is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String as = "" + tree.charAt(position);
   if (as.equals(pr))   counter++;
   if (as.equals(pl))  counter--;
   S = as + S;
   if (print2)
   System.out.println("Subtree = " + S);
   position = position -1;
   }
  String n = innerNode;
 //Replacement antiRule 2:
 //((S)n,Fn)i -> (S)i
 //where A and F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =  "("+ S  + n + "," + leaf + n + ")" + i ; 
 String	Si = S + i;
 if (print2)
  System.out.println("\nSnFni = " + SnFni + "\n Si = " + Si);
 //Rule 1 is executed
 newTree = tree.replace(SnFni, Si);	
 if (print1)
  System.out.println("Output tree = \n" + newTree);
 return newTree;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply   antirule3:
//(Fn,An)i  -> Ai
private static String  antiRule3(String tree, 
		int positionLeaf, int positionNode)
{
 if (print1)
  System.out.println("AntiRule3");
 int beginIndex = positionLeaf - 1;
 int endIndex = positionNode + 3;
 String FnAni = tree.substring(beginIndex, endIndex);
 if (print1)
  System.out.println("FnAni = " + FnAni);
 String Ai = "" + tree.charAt(beginIndex+4)
              + tree.charAt(endIndex-1);
 if (print1)
  System.out.println("Ai = " + Ai);
 tree = tree.replace(FnAni, Ai);	
 if (print1)
  System.out.println("Output tree = \n" + tree);
 return tree;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply antirule4:
//(Fn,(S)n)i -> (S)i
private static String  antiRule4(String tree, 
		int positionLeaf, int positionNode)
{
 if (print1)
  System.out.println("AntiRule4");
 //Leaf to be deleted	
 String leaf = "" + tree.charAt(positionLeaf);
 if (print1)
	 System.out.println("\nLeaf to be deleted = " + leaf
			 + " at position " + positionLeaf);
 //Accompanying inner node  
 String innerNode = "" 
		            + tree.charAt(positionNode);
 if (print1)
	 System.out.println("\nAccompanying node = " + innerNode
			 + " at position " + positionNode);
	
 int beginIndex = positionLeaf - 1;
 int endIndex = positionNode+2;
 String i = ""; //BUG?????
 if (endIndex < tree.length())
	         i = ""+ tree.charAt(endIndex);
	 
 if (print2)
	 System.out.println("EndIndex = " + 
	                      endIndex + " i = " + i);
	 
 String FnSni	 = tree.substring(beginIndex, endIndex) + i;
 if (print1)
	 System.out.println("FnSni = " + FnSni);
 String S = tree.substring(beginIndex+4, endIndex-2);
 if (print1)
 System.out.println("S = " + S);
 String	Si = S + i;
 if (print1)
	 System.out.println("Si = " + Si);
	 
 String newTree = tree.replace(FnSni, Si);	
 if (print1)
	 System.out.println("Output tree = \n" + newTree); 
	 
 return newTree;
}




//The last inserted  leaf is deleted from tree.
private static String deleteLastLeaf(String tree) 
{
 //Leaf to be deleted	
 String leaf = "" + leaves.charAt(nextLeafIndex);
 if (print1)
  System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
        + internalNodes.charAt(nextIntNodeIndex);
 if (print1)
  System.out.println("Accompanying inner  node = " 
                + innerNode);		
 //new tree = output
 String newTree = "";
 //length of  tree as string
 int l = tree.length();
 //Last positions are found
 int positionLeaf = -1;
 int positionNode = -1;  
 //tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {
	 String c = "" + tree.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
  }
 if (print1)
  {
   System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
   System.out.println("Position of " + innerNode + 
      " = " + positionNode);
  }
 int firstPosition = tree.indexOf(innerNode);
 //Check for the char before firstPosition	
char a = tree.charAt(firstPosition - 1);
 //If I links to a complex structure
 if (a == ')')
  newTree = antiRule2(tree,positionLeaf,positionNode);
 else  //Apply antiRule1 for simple cases
	  newTree = antiRule1(tree,positionLeaf,positionNode);
 nextLeafIndex--;
 nextIntNodeIndex--;
 return newTree;	
}


//A random leaf is deleted from tree. 
private static String deleteRandomLeaf(String tree) 
{
 if (print1)
 {
 System.out.println("\nMethod deleteRandomLeaf");
 System.out.println("deletableLeaves = " + deletableLeaves);
 }
 int l = deletableLeaves.length();
 int position = r.nextInt(l);
 //Leaf to be deleted	
 String leaf = "" + deletableLeaves.charAt(position);
 int nodeIndex = tree.indexOf(leaf) + 1;
 if (print1)
  System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
            + tree.charAt(nodeIndex);
 if (print1)
  System.out.println("Accompanying inner  node = " 
                    + innerNode);		
 //new tree = output
 String newTree = "";
 //length of  tree as string
 l = tree.length();
 //Last positions are found
 int positionLeaf = -1;
 int positionNode = -1;  
 //tree is scanned char by char
 for(int i = 0; i < l; i++ )
   {
	 String c = "" + tree.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
   }
 if (print1)
  {
   System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
   System.out.println("Position of " + innerNode + 
          " = " + positionNode);
  }
 //Which set of deleting rules to apply:
 // 1 and 2 else 3 and 4?
 String previousChar = "" + tree.charAt(positionLeaf-1);
 if (previousChar.equals(",")) //Apply antiRules 1 and 2
  {
   int firstPosition = tree.indexOf(innerNode);
   //Check for the char before firstPosition	
   char a = tree.charAt(firstPosition - 1);
   //If I links to a complex structure
   if (a == ')')
    newTree = antiRule2(tree,positionLeaf,positionNode);
    else  //Apply antiRule1 for simple cases
	  newTree = antiRule1(tree,positionLeaf,positionNode);
  nextLeafIndex--;
  nextIntNodeIndex--;
 }
 else //Apply antiRules 3 and 4
  {
   int firstPosition = tree.indexOf(innerNode);
   //Check for the char before firstPosition	
   char a = tree.charAt(firstPosition + 2);
   //If I links to a complex structure
   if (a == '(')
	 newTree = antiRule4(tree,positionLeaf,positionNode);
	    else  //Apply antiRule1 for simple cases
   newTree = antiRule3(tree,positionLeaf,positionNode);
   nextLeafIndex--;
   nextIntNodeIndex--;	 
  }
 deletableLeaves = delete(leaf,deletableLeaves);
  return newTree;	
}



//Mutation = 
//Random deletions are reinserted.
public static String  mutationRandomDeletion(
		           String tree, int nLeavesToBeInserted)
{
 //Some leaves are deleted: random order
 if (print1)
  System.out.println("Some leaves are deleted AT RANDOM");
 for(int i = 0; i < nLeavesToDelete; i++ )	 
		 {
		  //Last inserted leaf is deleted.
		  tree = deleteRandomLeaf(tree);
		 }
 //Deleted leaves are reinserted
 if (print1)
  System.out.println("Deleted leaves are reinserted");
 for(int i = 0; i < nLeavesToDelete; i++ )	 
	   {
		  tree = generateChild(tree);
	   }
 return tree;
}


//Mutation:
//some leaves are deleted and then reinserted 
//in random order.
//Deleting can happen in either of two ways:
//Reversing the order of synthesis else
//in random order.
public static String mutation(String topology, 
		                       int nLeavesToBeInserted)
{
 //Parental 3 leaves are not removed	
 deletableLeaves = presentLeaves.substring(3);
 //Number of leaves to be deleted and then reinserted
 nLeavesToDelete = r.nextInt(nLeavesToBeInserted-1) + 1;
 nLeavesToDelete = 4;
 //choose: random order else reverse to that of synthesis
 double pReverseSynth = 0;
 double s = r.nextDouble();
 if ( s < pReverseSynth)//Choose deletion is reversed order
  {
   //Some leaves are deleted: 
   //reversed order to that of synthesis.
   if (print1)
    System.out.println("Some leaves are deleted");
    for(int i = 0; i < nLeavesToDelete; i++ )	 
	 {
	  //Last inserted leaf is deleted.
	  topology = deleteLastLeaf(topology);
	  if (print1)
	   System.out.println(topology);
	 }
   //deleted leaves are reinserted
   if (print1)
    System.out.println("Deleted leaves are reinserted");
   for(int i = 0; i < nLeavesToDelete; i++ )	 
    {
	  topology = generateChild(topology);
    } 
  }
 else
  topology = mutationRandomDeletion(topology, nLeavesToDelete);
 //Output
 String mutant = topology;
 if (print1)
  System.out.println("\nOriginal and mutant topologies = \n" 
		  + original + "\n" + mutant);
 if (print1)
 if (original.equals(mutant))
			  System.out.println("Mutation made nothing");
	   else
			  System.out.println("\nMutation worked");	
 return mutant;
}


//Lengths are deleted from newTree
private static String deleteLengths(String newTree)
{
 String tree = "";
 String numeric = ".:0123456789";
 //length of string newTree
 int l = newTree.length();
 for(int i = 0; i < l; i++ )
  {
   //newTree is scanned char by char	 
   char c = newTree.charAt(i);
   String s = "" + c;
   //Is c numeric: delete it!
   if (numeric.contains(s)); //add char c to output
	  else tree = tree + c;
 }
 return tree;	 
}


//Tree undergoes mutation of topology and 
//also of lengths
private static String mutateTopology(String tree)
{
 String mutantTopology = "";
 String  mutantTree = "";
 //Topology is read from tree
 String topology = deleteLengths(tree);
 mutantTopology = mutation(topology, nLeavesToBeInserted);
 mutantTree = 	insertLengths(mutantTopology);

 mutantTree = mutateLengths(mutantTree);

return mutantTree;
}

//Random lengths are added to a tree in Link notation 
public static String insertLengths(String treeLinkN)
{
 String thisChar = "";
 String newTree = "";
 //length of String   tree
 int m = treeLinkN.length();
 //TreeLinkN is scanned char by char
 for(int i = 0; i < m; i++ )
  {	 
   thisChar = "" + treeLinkN.charAt(i); 
   newTree = newTree + thisChar;
   if (print1)
    System.out.println(newTree.length() + " " + newTree);
   //Lengths are appended
   if (leaves.contains(thisChar))
    {	
     int k = leaves.indexOf(thisChar);
     double j = lengthsLeaves[k];
    newTree = newTree + ":" + j;  	
    }
   if (internalNodes.contains(thisChar))
   {	
    int k = internalNodes.indexOf(thisChar);
    double j = lengthsInnerLinks[k];
 	 newTree = newTree + ":" + j;  
   }
  }
 return newTree;	  
}


//Input: string tree with complete information.
//Work: lengthsLeaves[] and lengthsInnerLinks[]
//are read from tree.
private static void decoderLengths(String tree)
{
 if(print2)	
  System.out.println(
	"Input tree at decoverLengths = \n "   + tree  );
 //Delete :
 tree = tree.replaceAll(":", "");
 if(print2)	
  System.out.println(
			"Input tree without : = \n "   + tree  );
 int n = tree.length();
 String thisChar = "";
 String nextChar = "";
 String numbers = ".0123456789";
 boolean leaf = false;
 boolean node = false;
 String Leaf = "";
 String numberLeaf = "";
 String numberNode = "";
 String Node = "";
 int j ;
 double f;

 //All chars in tree are scanned
 for(int i=0; i < n; i++)
 {
  thisChar = "" + tree.charAt(i);
  if (i < n-1) nextChar  = "" + tree.charAt(i+1);
  //Do we have a leaf?
  if (leaves.contains(thisChar)) 
   {
	 leaf = true; 
	 Leaf = "" + thisChar;
   }
 
 //Do we have an internal node?
 if (internalNodes.contains(thisChar)) 
  {
	 node = true;  
	 Node = "" + thisChar;
  }
 
 //Do we have a number that belongs in a leaf?
 if ( (leaf) & (numbers.contains(thisChar)))
	{
	 numberLeaf = numberLeaf + thisChar; 
	}

//Do we have a number that belongs in a leaf?
 if ( (node) & (numbers.contains(thisChar)))
	{
	 numberNode = numberNode + thisChar; 
	}
 
 //Must we end accumulation of numbers of a leaf?
 if ( (leaf) & (internalNodes.contains(nextChar)))
	{
	  leaf = false;
	  j = leaves.indexOf(Leaf);
	  if (print1)
	  System.out.println("Leaf = " + Leaf +
			  " numberLeaf = " + numberLeaf);
	  f = new Double(numberLeaf);
	  lengthsLeaves[j] = (int) Math.floor(f);
	  numberLeaf = "";
	}
 
 //Must we end accumulation of numbers of a node?
 if ( (node) & (!(numbers.contains(nextChar))))
	{
	  node = false;
	  j = internalNodes.indexOf(Node);
	  if (print1)
	  System.out.println("Node = " + Node);
	  f = new Double(numberNode);
	  lengthsInnerLinks[j] = (int) Math.floor(f);
	  numberNode = "";
	}
 
} 	 
}


//Tree undergoes mutation of lengths
private static String mutateLengths(String tree)
{
 String  mutantTree = "";
 //Topology is read from tree
 String topology = deleteLengths(tree);
 //Lengths are read from tree and recorded in vectors
 //lengthsLeaves[] and lengthsInnerLinks[]
 decoderLengths(tree);
 //Mutation in links that connect leaves
 for(int k = 0; k < numberOfLeaves; k++)
  {
	 double q = r.nextDouble();		
	 if (q < m) //mutate
			lengthsLeaves[k] = r.nextInt(maxLength);		 	
  }
 //Mutation in links that connect inner nodes
 for(int k = 0; k < numberOfLeaves-2; k++)
  {
	 double q = r.nextDouble();
	 if (q < m) //mutate
			lengthsInnerLinks[k] = r.nextInt(maxLength);		 	
  }
	
 mutantTree = 	insertLengths(topology);

 if (print1)
  {
   System.out.println("Input tree =  " + tree);
   System.out.println("Topology =  " + topology);
   System.out.println("Output tree= " + mutantTree);
  }

 return mutantTree;
}


//Type of  mutation is decided
private static String mutation(String tree)
{
 String  mutantTree = "";
 //Mutate lengths else topology ( + new lengths)?
 double s = r.nextDouble();
 if ( s < 0) 
	{
	 if (print1)
	 System.out.println("\n====Mutation of lengths");
	  mutantTree = mutateLengths(tree);
	 }
	else 
	 {
	 if (print1)
	  System.out.println("\n====Mutation of topology");
	  mutantTree = mutateTopology(tree);
	 }
	
 return mutantTree;
}

//Random lengths are added to a tree in Link notation 
public static void generateLengths()
{
 //Lengths with digit values are generated
 for(int i = 0; i < numberOfLeaves; i++)
 {
	 //Zero lengths are not allowed 
	 lengthsLeaves[i] = 1 + r.nextInt(maxLength - 1); 
	 lengthsInnerLinks[i] = 1 +  r.nextInt(maxLength - 1);
 }
} 


public static void main(String[] args) 
{
 //Are you going to work with the given order,
 //a reverse one, or with a permutation?
 double pGivenOrder = 1;
 double pReverseOrder = 0;
 double a = pGivenOrder + pReverseOrder;
 //Implicit: probability of permutation =
 //1 - (pGivenOrder+pReverseOrder).
 //Election is made
 double s = r.nextDouble();
 if ( s < pGivenOrder); //Do nothing = order is not changed
  else if ( s < a )	   leaves = reverse(leaves);
	 else leaves = permuteOrder(leaves);//permute if  s > a	 
 //Change of format
 System.out.println("Parent topology = \n" + parent
		 + " -> "    + simplify(parent)); 		  	
 //A topology topology is generated	

 String topology = parent;
 //Number of leaves to be inserted
 int nLeavesToBeInserted = 6; 
 for(int i = 0; i < nLeavesToBeInserted; i++ )	 
   topology = generateChild(topology);
 //This original topology topology is next subject to mutation
 original = topology;
 System.out.println("Original topology = " + original);
 String newTop = mutation(topology, nLeavesToBeInserted);
 generateLengths();
 String tree = insertLengths(newTop);
 System.out.println("Tree = \n" + tree);
 tree = mutation(tree);
 System.out.println("Mutant tree = \n" + tree);
}
}//End of main class K22 TopoMutation7